@@ -39,19 +39,28 @@
 
 #include "ojph_defs.h"
 #include "ojph_arch.h"
+#include "ojph_mem.h"
 #include "ojph_colour.h"
 #include "ojph_colour_local.h"
 
 namespace ojph {
+
+  // defined elsewhere
+  class line_buf;
+
   namespace local {
 
     //////////////////////////////////////////////////////////////////////////
-    void (*cnvrt_si32_to_si32_shftd)
-      (const si32 *sp, si32 *dp, int shift, ui32 width) = NULL;
+    void (*rev_convert)
+      (const line_buf *src_line, const ui32 src_line_offset, 
+       line_buf *dst_line, const ui32 dst_line_offset, 
+       si64 shift, ui32 width) = NULL;
 
     //////////////////////////////////////////////////////////////////////////
-    void (*cnvrt_si32_to_si32_nlt_type3)
-      (const si32* sp, si32* dp, int shift, ui32 width) = NULL;
+    void (*rev_convert_nlt_type3)
+      (const line_buf *src_line, const ui32 src_line_offset, 
+       line_buf *dst_line, const ui32 dst_line_offset, 
+       si64 shift, ui32 width) = NULL;
 
     //////////////////////////////////////////////////////////////////////////
     void (*cnvrt_si32_to_float_shftd)
@@ -71,13 +80,13 @@ namespace ojph {
 
     //////////////////////////////////////////////////////////////////////////
     void (*rct_forward)
-      (const si32 *r, const si32 *g, const si32 *b,
-       si32 *y, si32 *cb, si32 *cr, ui32 repeat) = NULL;
+      (const line_buf* r, const line_buf* g, const line_buf* b,
+       line_buf* y, line_buf* cb, line_buf* cr, ui32 repeat) = NULL;
 
     //////////////////////////////////////////////////////////////////////////
     void (*rct_backward)
-      (const si32 *y, const si32 *cb, const si32 *cr,
-       si32 *r, si32 *g, si32 *b, ui32 repeat) = NULL;
+      (const line_buf* r, const line_buf* g, const line_buf* b,
+       line_buf* y, line_buf* cb, line_buf* cr, ui32 repeat) = NULL;
 
     //////////////////////////////////////////////////////////////////////////
     void (*ict_forward)
@@ -100,8 +109,8 @@ namespace ojph {
 
 #if !defined(OJPH_ENABLE_WASM_SIMD) || !defined(OJPH_EMSCRIPTEN)
 
-      cnvrt_si32_to_si32_shftd = gen_cnvrt_si32_to_si32_shftd;
-      cnvrt_si32_to_si32_nlt_type3 = gen_cnvrt_si32_to_si32_nlt_type3;
+      rev_convert = gen_rev_convert;
+      rev_convert_nlt_type3 = gen_rev_convert_nlt_type3;
       cnvrt_si32_to_float_shftd = gen_cnvrt_si32_to_float_shftd;
       cnvrt_si32_to_float = gen_cnvrt_si32_to_float;
       cnvrt_float_to_si32_shftd = gen_cnvrt_float_to_si32_shftd;
@@ -130,10 +139,10 @@ namespace ojph {
       #ifndef OJPH_DISABLE_SSE2
         if (get_cpu_ext_level() >= X86_CPU_EXT_LEVEL_SSE2)
         {
+          rev_convert = sse2_rev_convert;
+          rev_convert_nlt_type3 = sse2_rev_convert_nlt_type3;
           cnvrt_float_to_si32_shftd = sse2_cnvrt_float_to_si32_shftd;
           cnvrt_float_to_si32 = sse2_cnvrt_float_to_si32;
-          cnvrt_si32_to_si32_shftd = sse2_cnvrt_si32_to_si32_shftd;
-          cnvrt_si32_to_si32_nlt_type3 = sse2_cnvrt_si32_to_si32_nlt_type3;
           rct_forward = sse2_rct_forward;
           rct_backward = sse2_rct_backward;
         }
@@ -154,8 +163,8 @@ namespace ojph {
       #ifndef OJPH_DISABLE_AVX2
         if (get_cpu_ext_level() >= X86_CPU_EXT_LEVEL_AVX2)
         {
-          cnvrt_si32_to_si32_shftd = avx2_cnvrt_si32_to_si32_shftd;
-          cnvrt_si32_to_si32_nlt_type3 = avx2_cnvrt_si32_to_si32_nlt_type3;
+          rev_convert = avx2_rev_convert;
+          rev_convert_nlt_type3 = avx2_rev_convert_nlt_type3;
           rct_forward = avx2_rct_forward;
           rct_backward = avx2_rct_backward;
         }
@@ -168,8 +177,9 @@ namespace ojph {
   #endif // !OJPH_DISABLE_SIMD
 
 #else // OJPH_ENABLE_WASM_SIMD
-      cnvrt_si32_to_si32_shftd = wasm_cnvrt_si32_to_si32_shftd;
-      cnvrt_si32_to_si32_nlt_type3 = wasm_cnvrt_si32_to_si32_nlt_type3;
+
+      rev_convert = wasm_rev_convert;
+      rev_convert_nlt_type3 = wasm_rev_convert_nlt_type3;
       cnvrt_si32_to_float_shftd = wasm_cnvrt_si32_to_float_shftd;
       cnvrt_si32_to_float = wasm_cnvrt_si32_to_float;
       cnvrt_float_to_si32_shftd = wasm_cnvrt_float_to_si32_shftd;
@@ -178,6 +188,7 @@ namespace ojph {
       rct_backward = wasm_rct_backward;
       ict_forward = wasm_ict_forward;
       ict_backward = wasm_ict_backward;
+
 #endif // !OJPH_ENABLE_WASM_SIMD
 
       colour_transform_functions_initialized = true;
@@ -201,20 +212,78 @@ namespace ojph {
 #if !defined(OJPH_ENABLE_WASM_SIMD) || !defined(OJPH_EMSCRIPTEN)
 
     //////////////////////////////////////////////////////////////////////////
-    void gen_cnvrt_si32_to_si32_shftd(const si32 *sp, si32 *dp, int shift,
-                                      ui32 width)
+    void gen_rev_convert(
+      const line_buf *src_line, const ui32 src_line_offset, 
+      line_buf *dst_line, const ui32 dst_line_offset, 
+      si64 shift, ui32 width)
     {
-      for (ui32 i = width; i > 0; --i)
-        *dp++ = *sp++ + shift;
+      if (src_line->flags & line_buf::LFT_32BIT)
+      { 
+        if (dst_line->flags & line_buf::LFT_32BIT)
+        {
+          const si32 *sp = src_line->i32 + src_line_offset;
+          si32 *dp = dst_line->i32 + dst_line_offset;
+          si32 s = (si32)shift;
+          for (ui32 i = width; i > 0; --i)
+            *dp++ = *sp++ + s;
+        }
+        else 
+        {
+          const si32 *sp = src_line->i32 + src_line_offset;
+          si64 *dp = dst_line->i64 + dst_line_offset;
+          for (ui32 i = width; i > 0; --i)
+            *dp++ = *sp++ + shift;
+        }
+      }
+      else 
+      {
+        assert(src_line->flags | line_buf::LFT_64BIT);
+        assert(dst_line->flags | line_buf::LFT_32BIT);
+        const si64 *sp = src_line->i64 + src_line_offset;
+        si32 *dp = dst_line->i32 + dst_line_offset;
+        for (ui32 i = width; i > 0; --i)
+          *dp++ = (si32)(*sp++ + shift);
+      }
     }
 
     //////////////////////////////////////////////////////////////////////////
-    void gen_cnvrt_si32_to_si32_nlt_type3(const si32 *sp, si32 *dp, 
-                                          int shift, ui32 width)
+    void gen_rev_convert_nlt_type3(
+      const line_buf *src_line, const ui32 src_line_offset, 
+      line_buf *dst_line, const ui32 dst_line_offset, 
+      si64 shift, ui32 width)
     {
-      for (ui32 i = width; i > 0; --i) {
-        const si32 v = *sp++;
-        *dp++ = v >= 0 ? v : (- v - shift);
+      if (src_line->flags & line_buf::LFT_32BIT)
+      { 
+        if (dst_line->flags & line_buf::LFT_32BIT)
+        {
+          const si32 *sp = src_line->i32 + src_line_offset;
+          si32 *dp = dst_line->i32 + dst_line_offset;
+          si32 s = (si32)shift;
+          for (ui32 i = width; i > 0; --i) {
+            const si32 v = *sp++;
+            *dp++ = v >= 0 ? v : (- v - s);
+          }
+        }
+        else 
+        {
+          const si32 *sp = src_line->i32 + src_line_offset;
+          si64 *dp = dst_line->i64 + dst_line_offset;
+          for (ui32 i = width; i > 0; --i) {
+            const si64 v = *sp++;
+            *dp++ = v >= 0 ? v : (- v - shift);
+          }
+        }
+      }
+      else 
+      {
+        assert(src_line->flags | line_buf::LFT_64BIT);
+        assert(dst_line->flags | line_buf::LFT_32BIT);
+        const si64 *sp = src_line->i64 + src_line_offset;
+        si32 *dp = dst_line->i32 + dst_line_offset;
+        for (ui32 i = width; i > 0; --i) {
+          const si64 v = *sp++;
+          *dp++ = (si32)(v >= 0 ? v : (- v - shift));
+        }
       }
     }
 
@@ -251,26 +320,104 @@ namespace ojph {
     }
 
     //////////////////////////////////////////////////////////////////////////
-    void gen_rct_forward(const si32 *r, const si32 *g, const si32 *b,
-                         si32 *y, si32 *cb, si32 *cr, ui32 repeat)
+    void gen_rct_forward(
+      const line_buf *r, const line_buf *g, const line_buf *b,
+      line_buf *y, line_buf *cb, line_buf *cr, ui32 repeat)
     {
-      for (ui32 i = repeat; i > 0; --i)
+      assert((y->flags  & line_buf::LFT_REVERSIBLE) &&
+             (cb->flags & line_buf::LFT_REVERSIBLE) && 
+             (cr->flags & line_buf::LFT_REVERSIBLE) &&
+             (r->flags  & line_buf::LFT_REVERSIBLE) &&
+             (g->flags  & line_buf::LFT_REVERSIBLE) && 
+             (b->flags  & line_buf::LFT_REVERSIBLE));
+      
+      if  (y->flags & line_buf::LFT_32BIT)
       {
-        *y++ = (*r + (*g << 1) + *b) >> 2;
-        *cb++ = (*b++ - *g);
-        *cr++ = (*r++ - *g++);
+        assert((y->flags  & line_buf::LFT_32BIT) &&
+               (cb->flags & line_buf::LFT_32BIT) && 
+               (cr->flags & line_buf::LFT_32BIT) &&
+               (r->flags  & line_buf::LFT_32BIT) &&
+               (g->flags  & line_buf::LFT_32BIT) && 
+               (b->flags  & line_buf::LFT_32BIT));        
+        const si32 *rp = r->i32, * gp = g->i32, * bp = b->i32;
+        si32 *yp = y->i32, * cbp = cb->i32, * crp = cr->i32;
+        for (ui32 i = repeat; i > 0; --i)
+        {
+          si32 rr = *rp++, gg = *gp++, bb = *bp++;
+          *yp++ = (rr + (gg << 1) + bb) >> 2;
+          *cbp++ = (bb - gg);
+          *crp++ = (rr - gg);
+        }
+      }
+      else 
+      {
+        assert((y->flags  & line_buf::LFT_64BIT) &&
+               (cb->flags & line_buf::LFT_64BIT) && 
+               (cr->flags & line_buf::LFT_64BIT) &&
+               (r->flags  & line_buf::LFT_32BIT) &&
+               (g->flags  & line_buf::LFT_32BIT) && 
+               (b->flags  & line_buf::LFT_32BIT));
+        const si32 *rp = r->i32, *gp = g->i32, *bp = b->i32;
+        si64 *yp = y->i64, *cbp = cb->i64, *crp = cr->i64;
+        for (ui32 i = repeat; i > 0; --i)
+        {
+          si64 rr = *rp++, gg = *gp++, bb = *bp++;
+          *yp++ = (rr + (gg << 1) + bb) >> 2;
+          *cbp++ = (bb - gg);
+          *crp++ = (rr - gg);
+        }
       }
     }
 
     //////////////////////////////////////////////////////////////////////////
-    void gen_rct_backward(const si32 *y, const si32 *cb, const si32 *cr,
-                          si32 *r, si32 *g, si32 *b, ui32 repeat)
+    void gen_rct_backward(
+      const line_buf *y, const line_buf *cb, const line_buf *cr,
+      line_buf *r, line_buf *g, line_buf *b, ui32 repeat)
     {
-      for (ui32 i = repeat; i > 0; --i)
+      assert((y->flags  & line_buf::LFT_REVERSIBLE) &&
+             (cb->flags & line_buf::LFT_REVERSIBLE) && 
+             (cr->flags & line_buf::LFT_REVERSIBLE) &&
+             (r->flags  & line_buf::LFT_REVERSIBLE) &&
+             (g->flags  & line_buf::LFT_REVERSIBLE) && 
+             (b->flags  & line_buf::LFT_REVERSIBLE));
+
+      if (y->flags & line_buf::LFT_32BIT)
       {
-        *g = *y++ - ((*cb + *cr)>>2);
-        *b++ = *cb++ + *g;
-        *r++ = *cr++ + *g++;
+        assert((y->flags  & line_buf::LFT_32BIT) &&
+               (cb->flags & line_buf::LFT_32BIT) && 
+               (cr->flags & line_buf::LFT_32BIT) &&
+               (r->flags  & line_buf::LFT_32BIT) &&
+               (g->flags  & line_buf::LFT_32BIT) && 
+               (b->flags  & line_buf::LFT_32BIT));
+        const si32 *yp = y->i32, *cbp = cb->i32, *crp = cr->i32;
+        si32 *rp = r->i32, *gp = g->i32, *bp = b->i32;
+        for (ui32 i = repeat; i > 0; --i)
+        {
+          si32 yy = *yp++, cbb = *cbp++, crr = *crp++;
+          si32 gg = yy - ((cbb + crr) >> 2);
+          *rp++ = crr + gg;
+          *gp++ = gg;
+          *bp++ = cbb + gg;
+        }
+      }
+      else
+      {
+        assert((y->flags  & line_buf::LFT_64BIT) &&
+               (cb->flags & line_buf::LFT_64BIT) && 
+               (cr->flags & line_buf::LFT_64BIT) &&
+               (r->flags  & line_buf::LFT_32BIT) &&
+               (g->flags  & line_buf::LFT_32BIT) && 
+               (b->flags  & line_buf::LFT_32BIT));   
+        const si64 *yp = y->i64, *cbp = cb->i64, *crp = cr->i64;
+        si32 *rp = r->i32, *gp = g->i32, *bp = b->i32;
+        for (ui32 i = repeat; i > 0; --i)
+        {
+          si64 yy = *yp++, cbb = *cbp++, crr = *crp++;
+          si64 gg = yy - ((cbb + crr) >> 2);
+          *rp++ = (si32)(crr + gg);
+          *gp++ = (si32)gg;
+          *bp++ = (si32)(cbb + gg);
+        }
       }
     }
 
